# -*- coding: utf-8 -*-
"""Classes and functions for manipulating/processing MSI information (status, data, locus, ...)."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2018 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.7.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import math
import json
import numpy as np
from copy import deepcopy


def toDict(msi_object):
    """
    Retuns a dictionary representing the object. This method is used in json.dump in argument named "default" for recursively convert an object to json.

    :param msi_object: The object to convert.
    :type msi_object: a class of anacore.msi library
    :return: The dictionary representing the object.
    :rtype: dict
    """
    return msi_object.__dict__


class Status:
    """Status for samples (MSISplRes) and loci (LocusRes) in anacore.msi library."""

    stable = "MSS"
    unstable = "MSI"
    undetermined = "Undetermined"  # Cannot be determined
    none = None  # Not evaluated

    @staticmethod
    def authorizedValues():
        """Return the values authorized for stability status."""
        return [attr_value for attr_name, attr_value in Status.__dict__.items() if attr_name != "authorizedValues" and not attr_name.startswith("__")]


class MSILocus:
    """Manage a locus of a microsatellite (name, position and stability status)."""

    def __init__(self, position, name=None, results=None):
        """
        Build and return an instance of MSILocus.

        :param position: The position of the locus with format chr:start-end (the start is 0-based).
        :type position: str
        :param name: The name of the locus (for example: NR21).
        :type name: str
        :param results: The results of the locus indexed by the name of the method used to produce these results.
        :type results: dict
        :return: The new instance.
        :rtype: MSILocus
        """
        self.position = position
        self.name = name
        self.results = {} if results is None else results

    def delResult(self, result_id):
        """
        Delete result from locus.

        :param result_id: The ID used for index the results to delete.
        :type result_id: str
        """
        self.results.pop(result_id, None)

    @staticmethod
    def fromDict(data):
        """
        Build and return an instance of MSILocus from a dict. This method is used for load instance from JSON.

        :param data: The locus information.
        :type data: dict
        :return: The new instance.
        :rtype: MSILocus
        """
        cleaned_data = deepcopy(data)
        if "results" in data:
            for method, result in data["results"].items():
                if "_class" in result and result["_class"] == "LocusResPairsCombi":
                    cleaned_data["results"][method] = LocusResPairsCombi.fromDict(result)
                elif "_class" in result and result["_class"] == "LocusResDistrib":
                    cleaned_data["results"][method] = LocusResDistrib.fromDict(result)
                else:
                    cleaned_data["results"][method] = LocusRes.fromDict(result)
        return MSILocus(**cleaned_data)


class LocusRes:
    """Manage the stability status for an anlysis of a locus."""

    def __init__(self, status, score=None, data=None):
        """
        Build and return an instance of LocusRes.

        :param status: The status of the locus.
        :type status: msi.Status
        :param score: The confidence score on the status prediction (0 to 1).
        :type score: float
        :param data: The data used to predict the status (example: the size distribution).
        :type results: dict
        :return: The new instance.
        :rtype: LocusRes
        """
        self._class = "LocusRes"
        self.status = status
        self.score = score
        self.data = {} if data is None else data

    @staticmethod
    def fromDict(data):
        """
        Build and return an instance of LocusRes from a dict. This method is used for load instance from JSON.

        :param data: The locus result information.
        :type data: dict
        :return: The new instance.
        :rtype: LocusRes
        """
        cleaned_data = deepcopy(data)
        if "_class" in cleaned_data:
            cleaned_data.pop("_class", None)
        return LocusRes(**cleaned_data)


class LocusResDistrib(LocusRes):
    """Manage the stability status for an anlysis of a locus containing the count by length."""

    def __init__(self, status, score=None, data=None):
        """
        Build and return an instance of LocusResDistrib.

        :param status: The status of the locus.
        :type status: msi.Status
        :param score: The confidence score on the status prediction (0 to 1).
        :type score: float
        :param data: The data used to predict the status (example: the size distribution).
        :type results: dict
        :return: The new instance.
        :rtype: LocusResDistrib
        """
        super().__init__(status, score, data)
        self._class = "LocusResDistrib"

    def getCount(self):
        """
        Return the number of elements in size distribution.

        :return: The number of elements in size distribution.
        :rtype: int
        """
        return sum(list(self.data["nb_by_length"].values()))

    def getMinLength(self):
        """
        Return the minimum length of the locus.

        :return: The minimum length of the locus.
        :rtype: int
        """
        return min([int(elt) for elt in self.data["nb_by_length"]])

    def getMaxLength(self):
        """
        Return the maximum length of the locus.

        :return: The maximum length of the locus.
        :rtype: int
        """
        return max([int(elt) for elt in self.data["nb_by_length"]])

    def getDensePrct(self, start=None, end=None):
        """
        Return the percentage of elements by locus length for absolutely all lengths betwen start and end. The lengths with 0 are also indicated. The percentage is based on all the distribution not reduced at start and end.

        :param start: The first length of the returned distribution. [Default: the minimum length in the distribution]
        :type start: int
        :param end: The last length of the returned distribution. [Default: the maximum length in the distribution]
        :type end: int
        :return: The number of elements in each length.
        :rtype: list
        """
        nb_pairs = self.getCount()
        dense_prct = list()
        for curr_count in self.getDenseCount(start, end):
            prct = None
            if nb_pairs != 0:
                prct = (curr_count * 100) / nb_pairs
            dense_prct.append(prct)
        return dense_prct

    def getDenseCount(self, start=None, end=None):
        """
        Return the number of elements by locus length for absolutely all lengths betwen start and end. The length with 0 are also indicated.

        :param start: The first length of the returned distribution. [Default: the minimum length in the distribution]
        :type start: int
        :param end: The last length of the returned distribution. [Default: the maximum length in the distribution]
        :type end: int
        :return: The number of elements in each length.
        :rtype: list
        """
        if start is None:
            start = self.getMinLength()
        if end is None:
            end = self.getMaxLength()
        dense_count = list()
        for curr_length in range(start, end + 1):
            count = 0
            if str(curr_length) in self.data["nb_by_length"]:
                count = self.data["nb_by_length"][str(curr_length)]
            dense_count.append(count)
        return dense_count

    @staticmethod
    def fromDict(data):
        """
        Build and return an instance of LocusResDistrib from a dict. This method is used for load instance from JSON.

        :param data: The locus result information.
        :type data: dict
        :return: The new instance.
        :rtype: LocusResDistrib
        """
        cleaned_data = deepcopy(data)
        if "_class" in cleaned_data:
            cleaned_data.pop("_class", None)
        return LocusResDistrib(**cleaned_data)


class LocusResPairsCombi(LocusResDistrib):
    """Manage the stability status for an anlysis after reads pair combination of a locus."""

    def __init__(self, status, score=None, data=None):
        """
        Build and return an instance of LocusResPairsCombi.

        :param status: The status of the locus.
        :type status: msi.Status
        :param score: The confidence score on the status prediction (0 to 1).
        :type score: float
        :param data: The data used to predict the status (example: the size distribution).
        :type results: dict
        :return: The new instance.
        :rtype: LocusResPairsCombi
        """
        super().__init__(status, score, data)
        self._class = "LocusResPairsCombi"

    def getNbFrag(self):
        """
        Return the number of fragments in size distribution.

        :return: The number of fragments in size distribution.
        :rtype: int
        """
        return self.getCount()

    @staticmethod
    def fromDict(data):
        """
        Build and return an instance of LocusResPairsCombi from a dict. This method is used for load instance from JSON.

        :param data: The locus result information.
        :type data: dict
        :return: The new instance.
        :rtype: LocusResPairsCombi
        """
        cleaned_data = deepcopy(data)
        if "_class" in cleaned_data:
            cleaned_data.pop("_class", None)
        return LocusResPairsCombi(**cleaned_data)


class MSISplRes:
    """Manage the stability status for an anlysis of a sample."""

    def __init__(self, status, score=None, method=None, param=None, version=None):
        """
        Build and return an instance of MSISplRes.

        :param status: The status of the sample.
        :type status: msi.Status
        :param score: The confidence score on the status prediction (0 to 1).
        :type score: float
        :param method: The name of the method used to predict status.
        :type method: str
        :param param: The parameters of the status prediction.
        :type param: dict
        :param version: The version of the method used to predict status.
        :type version: dict
        :return: The new instance.
        :rtype: MSISplRes
        """
        self.status = status
        self.method = method
        self.score = score
        self.param = param
        self.version = version

    @staticmethod
    def fromDict(data):
        """
        Build and return an instance of MSISplRes from a dict. This method is used for load instance from JSON.

        :param data: The sample result information.
        :type data: dict
        :return: The new instance.
        :rtype: MSISplRes
        """
        cleaned_data = deepcopy(data)
        return MSISplRes(**cleaned_data)


class MSISample:
    """Manage a sample in context of microsatellite instability analysis. This object contains information on loci and on analyses (at sample and loci levels)."""

    def __init__(self, name, loci=None, results=None):
        """
        Build and return an instance of MSISample.

        :param name: The sample name.
        :type name: str
        :param loci: By ID (format: chr:start-end with start 0-based) the loci targeted in analysis.
        :type loci: dict
        :param results: By method name the analyses of sample status.
        :type results: dict
        :return: The new instance.
        :rtype: MSISample
        """
        self.name = name
        self.loci = {} if loci is None else loci
        self.results = {} if results is None else results

    def getLociMethods(self):
        """
        Return the different methods used in loci status analysis.

        :return: The names of the methods used in analyses of loci.
        :rtype: set
        """
        methods = set()
        for curr_locus in self.loci:
            for curr_method in self.results.keys():
                methods.add(curr_method)
        return methods

    def addLocus(self, locus):
        """
        Add one locus to the sample.

        :param locus: The locus to add.
        :type locus: MSILocus
        """
        if locus.__class__.__name__ != "MSILocus":
            raise Exception('The class "{}" cannot be used as locus for MSISample.'.format(locus.__class__.__name__))
        self.loci[locus.position] = locus

    def delLoci(self, locus_ids):
        """
        Delete loci from the sample.

        :param locus_ids: The IDs of the loci to delete.
        :type locus_ids: list
        """
        for locus_id in locus_ids:
            self.delLocus(locus_id)

    def delLocus(self, locus_id):
        """
        Delete one locus from the sample.

        :param locus_id: The ID of the locus to delete.
        :type locus_id: str
        """
        self.loci.pop(locus_id, None)

    def _getStatusByMethod(self, method):
        """
        Return the list of loci status predicted by the selected method.

        :param method: The selected method.
        :type method: str
        :return: The list of status.
        :rtype: list
        """
        status = list()
        for locus_id, locus in self.loci.items():
            if method in locus.results:
                status.append(locus.results[method].status)
        return status

    def getNbUnstable(self, method):
        """
        Return the number of loci predicted as unstable with the selected method.

        :param method: The selected method.
        :type method: str
        :return: The number of unstable loci.
        :rtype: int
        """
        nb_unstable = 0
        status = self._getStatusByMethod(method)
        for curr_status in status:
            if curr_status is not None and curr_status == Status.unstable:
                nb_unstable += 1
        return nb_unstable

    def getNbStable(self, method):
        """
        Return the number of loci predicted as stable with the selected method.

        :param method: The selected method.
        :type method: str
        :return: The number of stable loci.
        :rtype: int
        """
        nb_stable = 0
        status = self._getStatusByMethod(method)
        for curr_status in status:
            if curr_status is not None and curr_status == Status.stable:
                nb_stable += 1
        return nb_stable

    def getNbUndetermined(self, method):
        """
        Return the number of loci where the status is undetermined after prediction with the selected method.

        :param method: The selected method.
        :type method: str
        :return: The number of undetermined loci.
        :rtype: int
        """
        nb_undetermined = 0
        status = self._getStatusByMethod(method)
        for curr_status in status:
            if curr_status is not None and curr_status == Status.undetermined:
                nb_undetermined += 1
        return nb_undetermined

    def getNbDetermined(self, method):
        """
        Return the number of loci with a prediction of status with the selected method.

        :param method: The selected method.
        :type method: str
        :return: The number of determined loci.
        :rtype: int
        """
        nb_determined = 0
        status = self._getStatusByMethod(method)
        for curr_status in status:
            if curr_status is not None and curr_status != Status.undetermined:
                nb_determined += 1
        return nb_determined

    def getNbProcessed(self, method):
        """
        Return the number of loci where the status has been evaluated for the sample with the selected method.

        :param method: The selected method.
        :type method: str
        :return: The number of processed loci.
        :rtype: int
        """
        nb_processed = 0
        status = self._getStatusByMethod(method)
        for curr_status in status:
            if curr_status is not None:
                nb_processed += 1
        return nb_processed

    def getNbLoci(self):
        """
        Return the number of loci in the sample.

        :param method: The selected method.
        :type method: str
        :return: The number of loci.
        :rtype: int
        """
        return len(self.loci)

    @staticmethod
    def fromDict(data):
        """
        Build and return an instance of MSISample from a dict. This method is used for load instance from JSON.

        :param data: The sample information.
        :type data: dict
        :return: The new instance.
        :rtype: MSISample
        """
        cleaned_data = deepcopy(data)
        # Loci
        if "loci" in data:
            cleaned_data["loci"] = dict()
            for locus_id, locus in data["loci"].items():
                cleaned_data["loci"][locus_id] = MSILocus.fromDict(locus)
        # Results
        if "results" in data:
            cleaned_data["results"] = dict()
            for method, result in data["results"].items():
                cleaned_data["results"][method] = MSISplRes.fromDict(result)
        # Name
        return MSISample(**cleaned_data)

    def _getScoreCalculation(self, eval_status, method, undetermined_weight=0.5, locus_weight_is_score=True):
        """
        Calculate and return a confidence score for the sample status prediction. This score is calculation take into account each locus and his score with the following formula: sum(scores) / (len(scores) + nb_loci_undetermined * undetermined_weight).

        :param eval_status: The score is calculated for this status (msi.Status.stable or msi.Status.unstable). Except for undetermined_weight equals to 0 the score of the complementary status CANNOT be calculated by 1 - complementary score.
        :type eval_status: msi.Status
        :param method: The status of the loci are extracted from the results of this method.
        :type method: str
        :param undetermined_weight: The weight of the undetermined loci in score calculation.
        :type undetermined_weight: float
        :param locus_weight_is_score: Use the prediction score of each locus as wheight of this locus in score calculation.
        :type locus_weight_is_score: bool
        :return: The prediction score. This score has a value between 0 and 1.
        :rtype: float
        """
        scores = list()
        nb_loci_undetermined = 0
        for locus_id, locus in self.loci.items():
            if locus.results[method].status is not None:
                if locus.results[method].status == Status.undetermined:
                    nb_loci_undetermined += 1
                else:
                    if locus.results[method].status == eval_status:
                        if locus_weight_is_score and locus.results[method].score is not None:
                            scores.append(locus.results[method].score)
                        else:
                            scores.append(1)
                    elif locus.results[method].status != Status.undetermined:
                        if locus_weight_is_score and locus.results[method].score is not None:
                            scores.append(1 - locus.results[method].score)
                        else:
                            scores.append(0)
        score = None
        if len(scores) != 0:
            score = sum(scores) / (len(scores) + nb_loci_undetermined * undetermined_weight)
        return round(score, 5)

    def setScore(self, method, undetermined_weight=0.5, locus_weight_is_score=True):
        """
        Calculate and set a confidence score for the sample status prediction. This score is calculation take into account each locus and his score with the following formula: sum(scores) / (len(scores) + nb_loci_undetermined * undetermined_weight).

        :param method: Name of the selected method.
        :type method: str
        :param undetermined_weight: The weight of the undetermined loci in score calculation.
        :type undetermined_weight: float
        :param locus_weight_is_score: Use the prediction score of each locus as wheight of this locus in score calculation.
        :type locus_weight_is_score: bool
        """
        spl_res = self.results[method]
        if spl_res.status == Status.undetermined or spl_res.status == Status.none:
            spl_res.score = None
        else:
            spl_res.score = self._getScoreCalculation(spl_res.status, method, undetermined_weight)

    def setStatusByInstabilityCount(self, method, min_voting_loci=3, instability_threshold=3):
        """
        Add status and score for the sample by number of unstable loci for the selected method.

        :param method: Name of the selected method.
        :type method: str
        :param min_voting_loci: Minimum number of voting loci (stable + unstable) to determine the sample status. If the number of voting loci is lower than this value the status for the sample will be undetermined.
        :type min_voting_loci: int
        :param instability_threshold: If the number of unstable is superior or equal than this value the status of the sample will be unstable.
        :type instability_threshold: int
        """
        result = MSISplRes.fromDict({
            'status': Status.undetermined,
            'method': method,
            'score': None,
            'param': {"aggregation_method": "instability count", "min_voting_loci": min_voting_loci, "instability_threshold": instability_threshold},
            'version': "1.0.0"
        })
        nb_stable = self.getNbStable(method)
        nb_unstable = self.getNbUnstable(method)
        result.status = Status.undetermined
        if nb_stable + nb_unstable >= min_voting_loci:
            if nb_unstable >= instability_threshold:
                result.status = Status.unstable
            else:
                result.status = Status.stable
        self.results[method] = result
        self.setScore(method)

    def setStatusByInstabilityRatio(self, method, min_voting_loci=1, instability_threshold=0.4):
        """
        Add status and score for the sample by ratio of unstable loci for the selected method.

        :param method: Name of the selected method.
        :type method: str
        :param min_voting_loci: Minimum number of voting loci (stable + unstable) to determine the sample status. If the number of voting loci is lower than this value the status for the sample will be undetermined.
        :type min_voting_loci: int
        :param instability_threshold: If the ratio unstable/(stable + unstable) is superior or equal than this value the status of the sample will be unstable.
        :type instability_threshold: float
        """
        result = MSISplRes.fromDict({
            'status': Status.undetermined,
            'method': method,
            'score': None,
            'param': {"aggregation_method": "instability ratio", "min_voting_loci": min_voting_loci, "instability_threshold": instability_threshold},
            'version': "1.0.0"
        })
        nb_stable = self.getNbStable(method)
        nb_unstable = self.getNbUnstable(method)
        result.status = Status.undetermined
        if nb_stable + nb_unstable >= min_voting_loci:
            if nb_unstable / (nb_stable + nb_unstable) >= instability_threshold:
                result.status = Status.unstable
            else:
                result.status = Status.stable
        self.results[method] = result
        self.setScore(method)

    def setStatusByMajority(self, method, min_voting_loci=1):
        """
        Add status and score for the sample by consensus on loci results for the selected method.

        :param method: Name of the selected method.
        :type method: str
        :param min_voting_loci: Minimum number of voting loci (stable + unstable) to determine the sample status. If the number of voting loci is lower than this value the status for the sample will be undetermined.
        :type min_voting_loci: int
        """
        result = MSISplRes.fromDict({
            'status': Status.undetermined,
            'method': method,
            'score': None,
            'param': {"aggregation_method": "majority", "min_voting_loci": min_voting_loci},
            'version': "1.0.0"
        })
        nb_stable = self.getNbStable(method)
        nb_unstable = self.getNbUnstable(method)
        if nb_stable + nb_unstable >= min_voting_loci:
            result.status = Status.undetermined
            if nb_stable > nb_unstable:
                result.status = Status.stable
            elif nb_stable < nb_unstable:
                result.status = Status.unstable
        self.results[method] = result
        self.setScore(method)


class LocusClassifier:
    """
    Classifier for locus using MSISample objects.

    Synopsis:
        clf = LocusClassifier(locus_id, method_name, classifier)
        clf.fit(train_dataset)
        clf.predict(test_dataset)
        clf.predict_proba(test_dataset)


        clf = LocusClassifier(locus_id, method_name, classifier)
        clf.fit(train_dataset)
        clf.set_status(test_dataset)
    """

    def __init__(self, locus_id, method_name, classifier, model_method_name="model", data_method_name=None):
        """
        Build and return an instance of LocusClassifier.

        :param locus_id: The ID (format: chr:start-end with start 0-based) of the loci targeted in analysis.
        :type locus_id: str
        :param method_name: By method name the analyses of sample status.
        :type method_name: dict
        :param classifier: The classifier used to predict status.
        :type classifier: A sklearn classifier object
        :param model_method_name: The name of the method used in models loci to store expected status and lengths distributions.
        :type model_method_name: str
        :param data_method_name: The data used for the prediction are extracted from the results of this method. By default the selected method is the same of the classifier method_name.
        :type data_method_name: str
        :return: The new instance.
        :rtype: LocusClassifier
        """
        self.classifier = classifier
        self.data_method_name = method_name if data_method_name is None else data_method_name
        self.locus_id = locus_id
        self.method_name = method_name
        self.model_method_name = model_method_name
        self._max_len = None  # This value is used to uniformise lengths distributions in the comparison
        self._min_len = None  # This value is used to uniformise lengths distributions in the comparison
        self._test_dataset = []
        self._train_dataset = []  # The MSISample provided for fitting. These samples are filtered before fit (see _usable_train_dataset)
        self._usable_train_dataset = []  # The MSISample containing status information for the locus

    def _get_min_max_len(self, dataset, method):
        """
        Return the minimum and maximum length for the locus in the dataset.

        :param dataset: The list of MSISample.
        :type dataset: list
        :param method: The lengths distribution of the loci are extracted from the results of this method.
        :type method: str
        :return: The list of MSISample.
        :rtype: list
        """
        min_len = math.inf
        max_len = -1
        for curr_spl in dataset:
            locus_res = curr_spl.loci[self.locus_id].results[method]
            min_len = min(min_len, locus_res.getMinLength())
            max_len = max(max_len, locus_res.getMaxLength())
        return min_len, max_len

    def _set_min_max_len(self):
        """Set the minimum and maximum length for the locus in the dataset (train + test)."""
        train_min, train_max = self._get_min_max_len(self._usable_train_dataset, self.model_method_name)
        test_min, test_max = self._get_min_max_len(self._test_dataset, self.data_method_name)
        self._min_len = min(train_min, test_min)
        self._max_len = max(train_max, test_max)

    def _get_data(self, dataset, method):
        """
        Return uniformised lengths distribution from the dataset.

        :param dataset: The list of MSISample.
        :type dataset: list
        :param method: The lengths distribution of the loci are extracted from the results of this method.
        :type method: str
        :return: The uniformised lengths distribution. Rows are samples, columns are lengths and values are percentages of counts in the sample.
        :rtype: np.matrix
        """
        prct_matrix = []  # rows are samples, columns are lengths and values are percentages of counts in the sample.
        if self._min_len is None and self._max_len is None:
            self._set_min_max_len()
        for curr_spl in dataset:
            locus_res = curr_spl.loci[self.locus_id].results[method]
            prct_matrix.append(
                locus_res.getDensePrct(self._min_len, self._max_len)
            )
        return np.matrix(prct_matrix)

    def _get_test_data(self):
        """
        Return uniformised lengths distribution for samples in test dataset.

        :return: The uniformised lengths distribution. Rows are samples, columns are lengths and values are percentages of counts in the sample.
        :rtype: np.matrix
        """
        return self._get_data(self._test_dataset, self.data_method_name)

    def _get_train_data(self):
        """
        Return uniformised lengths distribution for samples in usable train dataset.

        :return: The uniformised lengths distribution. Rows are samples, columns are lengths and values are percentages of counts in the sample.
        :rtype: np.matrix
        """
        return self._get_data(self._usable_train_dataset, self.model_method_name)

    def _get_train_labels(self):
        """
        Return the list of labels for samples in usable train dataset.

        :return: The list of labels for samples in usable train dataset.
        :rtype: np.array
        """
        labels = []
        for curr_spl in self._usable_train_dataset:
            locus_res = curr_spl.loci[self.locus_id].results[self.model_method_name]
            labels.append(locus_res.status)
        return np.array(labels)

    def fit(self, train_dataset):
        """
        Fit the model using train_dataset as training data and their status as target values.

        :param train_dataset: The list of MSISample containing the locus to classify and in LocusRes the data of the selected method and the status ecpected.
        :type test_dataset: list
        """
        self._train_dataset = train_dataset
        self._usable_train_dataset = [spl for spl in train_dataset if self.model_method_name in spl.loci[self.locus_id].results]
        self.classifier.fit(self._get_train_data(), self._get_train_labels())

    def predict(self, test_dataset):
        """
        Predict the status of the locus in the list of samples.

        :param test_dataset: The list of MSISample containing the locus to classify and in LocusRes the data of the selected method. These samples must already be filtered, for eaxmple on count of elements in length distribution.
        :type test_dataset: list
        :return: The predicted class for the selected locus in each sample.
        :rtype: np.array
        """
        self._test_dataset = test_dataset
        test_min_len, test_max_len = self._get_min_max_len(self._test_dataset, self.data_method_name)
        if test_min_len < self._min_len or test_max_len > self._max_len:
            self.fit(self._train_dataset)
        return self.classifier.predict(self._get_test_data())

    def predict_proba(self, test_dataset):
        """
        Return estimated probabilities for the test_dataset.

        :param test_dataset: The list of MSISample containing the locus to classify and in LocusRes the data of the selected method. These samples must already be filtered, for eaxmple on count of elements in length distribution.
        :type test_dataset: list
        :return: The estimated probabilities for the test_dataset.
        :rtype: np.array
        """
        self._test_dataset = test_dataset
        return self.classifier.predict_proba(self._get_test_data())

    def _get_scores(self, pred_labels):
        """
        Return confidence scores for the predictions of the test_dataset.

        :param pred_labels: The list of predicted status for the test_dataset.
        :type pred_labels: list
        :return: The scores for the predictions. The values are between 0 and 1.
        :rtype: list
        """
        scores = None
        proba_idx_by_label = {label: idx for idx, label in enumerate(self.classifier.classes_)}
        try:
            proba = self.predict_proba(self._test_dataset)
            scores = [round(spl_proba[proba_idx_by_label[spl_label]], 6) for spl_proba, spl_label in zip(proba, pred_labels)]
        except Exception:
            scores = [None for spl in self._test_dataset]
        return scores

    def set_status(self, test_dataset):
        """
        Set status and score for the selected locus in the list of samples.

        :param test_dataset: The list of MSISample containing the locus to classify and in LocusRes the data of the selected method. These samples must already be filtered, for eaxmple on count of elements in length distribution.
        :type test_dataset: list
        """
        self._test_dataset = test_dataset
        pred_labels = self.predict(test_dataset)
        pred_scores = self._get_scores(pred_labels)
        for label, score, sample in zip(pred_labels, pred_scores, self._test_dataset):
            if self.method_name not in sample.loci[self.locus_id].results:  # If the method does not exist in locus results
                sample.loci[self.locus_id].results[self.method_name] = LocusRes(Status.none)
            locus_res = sample.loci[self.locus_id].results[self.method_name]
            if self.method_name != self.data_method_name:  # If data used in prediction come from an other method
                locus_res.data["data_source"] = self.data_method_name
            locus_res.status = label
            locus_res.score = score


class MSIReport:
    """Read/write the JSON file used to store a list of MSISample."""

    @staticmethod
    def parse(in_path):
        """
        Return the list of MSISample stored in a MSIReport file.

        :param in_path: Path to the file storing MSISamples (format: MSIReport).
        :type in_path: str
        :return: The list of MSISample.
        :rtype: list
        """
        spl_data = []
        with open(in_path) as FH_in:
            spl_data = json.load(FH_in)
        return [MSISample.fromDict(curr_spl) for curr_spl in spl_data]

    @staticmethod
    def write(msi_samples, out_path):
        """
        Write the list of MSISample in a MSIReport file.

        :param msi_samples: The list of MSISample.
        :type msi_samples: list
        :param out_path: Path to the output file storing MSISamples (format: MSIReport).
        :type in_path: str
        """
        with open(out_path, "w") as FH_out:
            json.dump(msi_samples, FH_out, sort_keys=True, default=toDict)


def getNbSupporting(report, method="model", loci=None):
    """
    Return the number of samples by locus class.

    :param in_report: List of MSISample.
    :type in_report: list()
    :param method: Evaluated method.
    :type method: str
    :param loci: List of evaluated loci.
    :type loci: list
    :return: The number of samples by locus class. Each item is dict with following keys: locus_name, locus_id, status and support.
    :rtype: list
    """
    # Init loci from report if the argument is None
    if loci is None:
        loci = set()
        for spl in report:
            for locus_id in spl.loci:
                loci.add(locus_id)
    # Get count by status by locus
    nb_by_locus = {}
    for spl in report:
        for locus_id in loci:
            locus = spl.loci[locus_id]
            if locus_id not in nb_by_locus:
                nb_by_locus[locus_id] = {
                    "locus_name": locus.name,
                    "supp_by_status": {elt: 0 for elt in Status.authorizedValues()}
                }
            if method in locus.results:
                locus_status = locus.results[method].status
                nb_by_locus[locus_id]["supp_by_status"][locus_status] += 1
    # To list
    counts = []
    for locus_id, locus_info in nb_by_locus.items():
        for status, support in locus_info["supp_by_status"].items():
            counts.append({
                "locus_id": locus_id,
                "locus_name": locus_info["locus_name"],
                "status": status,
                "support": support
            })
    return counts


def getIncompleteModels(in_report, min_support_model=20):
    """
    Return the list of locus class model with an unsufficient number of supporting samples.

    :param in_report: Path to a MSIReport file.
    :type in_report: str
    :param min_support_model: The minimum number of sample in locus class to validate the model.
    :type min_support_model: int
    :return: The list of locus class model with with an unsufficient number of supporting samples. Each item is dict with following keys: locus_name, locus_id, status and support.
    :rtype: list
    """
    incomplete_models = []
    report = MSIReport.parse(in_report)
    for curr_model in getNbSupporting(report, method="model"):
        if curr_model["status"] in [Status.stable, Status.unstable] and curr_model["support"] < min_support_model:
            incomplete_models.append(curr_model)
    return incomplete_models



#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2020 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.1.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import sys
import unittest

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.hgvs import HGVSProtChange


########################################################################
#
# FUNCTIONS
#
########################################################################
class TestHGVSProtChange(unittest.TestCase):
    def testInsCouldBeIdentical(self):
        dataset = {
            "Ala3_Tyr4insAla": {
                "Ala3Dup": True
            },
            "T16_I17insGTTT": {
                "L16V": False,
                "G13_T16dup": True,
                "G13_L16dup": False,
                "G13_L16[1]": False,
                "G13_L16[2]": False
            },
            "L16_I17insGTTL": {
                "L16V": False,
                "G13_T16dup": False,
                "G13_L16dup": True,
                "G13_L16[1]": True,
                "G13_L16[2]": False
            },
            "L16_I17insGTTLGTTL": {
                "L16V": False,
                "G13_T16dup": False,
                "G13_L16dup": False,
                "G13_L16[1]": False,
                "G13_L16[2]": True
            },
            "L16_I17insGTTLGTLL": {
                "L16V": False,
                "G13_T16dup": False,
                "G13_L16dup": False,
                "G13_L16[1]": False,
                "G13_L16[2]": False
            }
        }
        for hgvs_ins, expected_by_cmp in dataset.items():
            hgvs_ins_obj = HGVSProtChange.fromStr(hgvs_ins)
            for hgvs_repeat, expected in expected_by_cmp.items():
                hgvs_repeat_obj = HGVSProtChange.fromStr(hgvs_repeat)
                self.assertEqual(HGVSProtChange.insCouldBeIdentical(hgvs_ins_obj, hgvs_repeat_obj), expected)

    def testIsInFrameIns(self):
        dataset = {
            # Special cases
            "?": False,
            "0": False,
            # Substitution
            "Val600=": False,
            "Val600Glu": False,
            "(G56A^S^C)": False,  # Uncertain
            "(Gly56Ala^Ser^Cys)": False,  # Uncertain
            "W24=/C": False,  # Mosaic
            "Trp24=/Cys": False,  # Mosaic
            # Deletion
            "(Val7del)": False,
            "Trp4del": False,
            "Lys23_Val25del": False,
            "(Pro458_Gly460del)": False,
            "Gly2_Met46del": False,
            "Trp26*": False,
            "Pro551_Glu554delProMetTyrGlu": False,
            "Trp557_Lys558DelTrpLys": False,
            "Asp842_His845DelAspIleMetHis": False,
            "Asp842DelAsp": False,
            # DelIns
            "Cys28delinsTrpVal": False,
            # Duplication
            "Ala3dup": True,
            "Ala3_Ser5dup": True,
            "A767_V769dupASV": True,
            # Insertion
            "His4_Gln5insAla": True,
            "Lys2_Gly3insGlnSerLys": True,
            "(Met3_His4insGlyTer)": True,
            "Arg78_Gly79ins23": True,
            "Gln746_Lys747ins*63": True,
            "(Ser332_Ser333ins(1))": True,
            "(Ser332_Ser333insX)": True,
            "(Val582_Asn583ins(5))": True,
            "(Val582_Asn583insXaaXaaXaaXaaXaa)": True,
            # Frameshift
            "Arg97ProfsTer23": False,
            "Arg97fs": False,
            "(Tyr4*)": False,
            "Ile327Argfs*?": False,
            "Gln151Thrfs*9": False,
            "*757Leu": False,
            "Val600Ter": False,
            # Extension
            "Met1ext-5": False,
            "Ter110GlnextTer17": False,
            "(Ter315TyrextAsnLysGlyThrTer)": False,
            "*315TyrextAsnLysGlyThr*": False,
            "Ter327Argext*?": False,
            # Reapeated sequences
            "Ala2[10]": True,
            "Ala2[1]": True,
            "Ala2_Pro5[10]": True,
            "Ala2_Pro5[1]": True,
        }
        for hgvs, expected in dataset.items():
            hgvs_obj = HGVSProtChange.fromStr(hgvs)
            self.assertEqual(hgvs_obj.isInFrameIns(), expected)

    def testParse(self):
        expected = HGVSProtChange("Ter", 757, None, None, None, None, ["Leu"], False)
        observed = HGVSProtChange.fromStr("*757L")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("*757Leu")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Val", 600, None, None, None, None, ["="], False)
        observed = HGVSProtChange.fromStr("V600=")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Val600=")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Val", 600, None, None, None, None, ["Glu"], False)
        observed = HGVSProtChange.fromStr("V600E")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Val600Glu")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Val", 600, None, None, None, None, ["Ter"], False)
        observed = HGVSProtChange.fromStr("V600*")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Val600Ter")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Val", 600, None, None, None, "del", [], False)
        observed = HGVSProtChange.fromStr("V600del")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Val600del")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Val", 600, "Ala", 603, None, "del", [], False)
        observed = HGVSProtChange.fromStr("V600_A603del")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Val600_Ala603del")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Val", 600, "Ala", 603, None, "dup", [], False)
        observed = HGVSProtChange.fromStr("V600_A603dup")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Val600_Ala603dup")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Val", 600, "Val", 601, None, "ins", ["(15)"], False)
        observed = HGVSProtChange.fromStr("Val600_Val601ins(15)")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Val", 600, "Val", 601, None, "ins", ["15"], False)
        observed = HGVSProtChange.fromStr("Val600_Val601ins15")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Cys", 28, None, None, None, "delins", ["Trp", "Val"], False)
        observed = HGVSProtChange.fromStr("C28delinsWV")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Cys28delinsTrpVal")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Cys", 28, "Lys", 29, None, "delins", ["Trp", "Val", "Ter"], False)
        observed = HGVSProtChange.fromStr("C28_K29delinsWV*")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Cys28_Lys29delinsTrpValTer")
        self.assertEqual(observed, expected)
        # Special cases
        expected = HGVSProtChange(None, None, None, None, None, None, None, False)
        observed = HGVSProtChange.fromStr("?")  # Unknown
        self.assertEqual(observed, expected)
        expected = HGVSProtChange(None, "0", None, None, None, None, None, False)
        observed = HGVSProtChange.fromStr("0")  # No protein
        self.assertEqual(observed, expected)
        # Substitution
        expected = HGVSProtChange("Trp", 24, None, None, None, None, ["Cys"], False)
        observed = HGVSProtChange.fromStr("W24C")  # Missense
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Trp24Cys")  # Missense
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Trp", 24, None, None, None, None, ["Cys"], True)
        observed = HGVSProtChange.fromStr("(Trp24Cys)")  # Missense
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Trp", 24, None, None, None, None, ["Ter"], False)
        observed = HGVSProtChange.fromStr("W24*")  # Nonsense
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Trp24*")  # Nonsense
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Trp24Ter")  # Nonsense
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Cys", 188, None, None, None, None, ["="], False)
        observed = HGVSProtChange.fromStr("C188=")  # No change
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Cys188=")  # No change
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Gly", 56, None, None, None, None, ["Ala", "^", "Ser", "^", "Cys"], True)
        observed = HGVSProtChange.fromStr("(G56A^S^C)")  # Uncertain
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("(Gly56Ala^Ser^Cys)")  # Uncertain
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Trp", 24, None, None, None, None, ["=", "/", "Cys"], False)
        observed = HGVSProtChange.fromStr("W24=/C")  # Mosaic
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Trp24=/Cys")  # Mosaic
        self.assertEqual(observed, expected)
        # Deletion
        expected = HGVSProtChange("Val", 7, None, None, None, "del", [], False)
        observed = HGVSProtChange.fromStr("V7del")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Val7del")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Val", 7, None, None, None, "del", [], True)
        observed = HGVSProtChange.fromStr("(V7del)")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("(Val7del)")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Trp", 4, None, None, None, "del", [], False)
        observed = HGVSProtChange.fromStr("Trp4del")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Lys", 23, "Val", 25, None, "del", [], False)
        observed = HGVSProtChange.fromStr("K23_V25del")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Lys23_Val25del")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Pro", 458, "Gly", 460, None, "del", [], True)
        observed = HGVSProtChange.fromStr("(P458_G460del)")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("(Pro458_Gly460del)")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Gly", 2, "Met", 46, None, "del", [], False)
        observed = HGVSProtChange.fromStr("G2_M46del")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Gly2_Met46del")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Trp", 26, None, None, None, None, ["Ter"], False)
        observed = HGVSProtChange.fromStr("W26*")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Trp26*")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Pro", 551, "Glu", 554, None, "del", [], False)
        observed = HGVSProtChange.fromStr("P551_E554delPMYE")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Pro551_Glu554delProMetTyrGlu")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Trp", 557, "Lys", 558, None, "del", [], False)
        observed = HGVSProtChange.fromStr("W557_K558DELWK")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Trp557_Lys558DelTrpLys")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Asp", 842, "His", 845, None, "del", [], False)
        observed = HGVSProtChange.fromStr("D842_H845DelDIMH")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Asp842_His845DelAspIleMetHis")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Asp", 842, None, None, None, "del", [], False)
        observed = HGVSProtChange.fromStr("D842DELD")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Asp842DelAsp")
        self.assertEqual(observed, expected)
        # observed = HGVSProtChange.fromStr("Val7=/del")
        # Duplication
        expected = HGVSProtChange("Ala", 767, "Val", 769, None, "dup", [], False)
        observed = HGVSProtChange.fromStr("A767_V769dupASV")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ala", 3, None, None, None, "dup", [], False)
        observed = HGVSProtChange.fromStr("Ala3dup")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("A3dup")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ala", 3, "Ser", 5, None, "dup", [], False)
        observed = HGVSProtChange.fromStr("A3_S5dup")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Ala3_Ser5dup")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ser", 6, None, None, None, "dup", [], False)
        observed = HGVSProtChange.fromStr("S6dup")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Ser6dup")
        self.assertEqual(observed, expected)
        # Insertion
        expected = HGVSProtChange("His", 4, "Gln", 5, None, "ins", ["Ala"], False)
        observed = HGVSProtChange.fromStr("H4_Q5insA")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("His4_Gln5insAla")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Lys", 2, "Gly", 3, None, "ins", ["Gln", "Ser", "Lys"], False)
        observed = HGVSProtChange.fromStr("K2_G3insQSK")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Lys2_Gly3insGlnSerLys")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Met", 3, "His", 4, None, "ins", ["Gly", "Ter"], True)
        observed = HGVSProtChange.fromStr("(M3_H4insG*)")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("(Met3_His4insGlyTer)")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Arg", 78, "Gly", 79, None, "ins", ["23"], False)
        observed = HGVSProtChange.fromStr("R78_G79ins23")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Arg78_Gly79ins23")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Gln", 746, "Lys", 747, None, "ins", ["Ter", "63"], False)
        observed = HGVSProtChange.fromStr("Q746_K747ins*63")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Gln746_Lys747ins*63")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ser", 332, "Ser", 333, None, "ins", ["(1)"], True)
        observed = HGVSProtChange.fromStr("(S332_S333ins(1))")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("(Ser332_Ser333ins(1))")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ser", 332, "Ser", 333, None, "ins", ["Xaa"], True)
        observed = HGVSProtChange.fromStr("(S332_S333insX)")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("(Ser332_Ser333insX)")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Val", 582, "Asn", 583, None, "ins", ["(5)"], True)
        observed = HGVSProtChange.fromStr("(V582_N583ins(5))")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("(Val582_Asn583ins(5))")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Val", 582, "Asn", 583, None, "ins", ["Xaa", "Xaa", "Xaa", "Xaa", "Xaa"], True)
        observed = HGVSProtChange.fromStr("(Val582_Asn583insXXXXX)")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("(Val582_Asn583insXaaXaaXaaXaaXaa)")
        self.assertEqual(observed, expected)
        # Frameshift
        expected = HGVSProtChange("Val", 2288, None, None, None, "fs", ["Ter", "1"], False)
        observed = HGVSProtChange.fromStr("V2288FS*1")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Arg", 97, None, None, "Pro", "fs", ["Ter", "23"], False)
        observed = HGVSProtChange.fromStr("R97Pfs*23")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Arg97ProfsTer23")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Arg", 97, None, None, None, "fs", [], False)
        observed = HGVSProtChange.fromStr("R97FS")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Arg97fs")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Tyr", 4, None, None, None, None, ["Ter"], True)
        observed = HGVSProtChange.fromStr("(Y4*)")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("(Tyr4*)")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ile", 327, None, None, "Arg", "fs", ["Ter", "?"], False)
        observed = HGVSProtChange.fromStr("I327Rfs*?")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Ile327Argfs*?")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Gln", 151, None, None, "Thr", "fs", ["Ter", "9"], False)
        observed = HGVSProtChange.fromStr("Q151TFS*9")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Gln151Thrfs*9")
        self.assertEqual(observed, expected)
        # Extension
        expected = HGVSProtChange("Met", 1, None, None, None, "ext", ["-5"], False)
        observed = HGVSProtChange.fromStr("M1ext-5")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Met1ext-5")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ter", 110, None, None, "Gln", "ext", ["Ter", "17"], False)
        observed = HGVSProtChange.fromStr("*110Qext*17")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Ter110GlnextTer17")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("*110Glnext*17")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ter", 315, None, None, "Tyr", "ext", ["Asn", "Lys", "Gly", "Thr", "Ter"], True)
        observed = HGVSProtChange.fromStr("(*315YextNKGT*)")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("(Ter315TyrextAsnLysGlyThrTer)")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ter", 315, None, None, "Tyr", "ext", ["Asn", "Lys", "Gly", "Thr", "Ter"], False)
        observed = HGVSProtChange.fromStr("*315TyrextAsnLysGlyThr*")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ter", 327, None, None, "Arg", "ext", ["Ter", "?"], False)
        observed = HGVSProtChange.fromStr("*327Rext*?")
        self.assertEqual(observed, expected)
        observed = HGVSProtChange.fromStr("Ter327Argext*?")
        self.assertEqual(observed, expected)
        # Reapeated sequences
        expected = HGVSProtChange("Ala", 2, None, None, None, None, ["[", "10", "]"], False)
        observed = HGVSProtChange.fromStr("A2[10]")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ala", 2, None, None, None, None, ["[", "10", "]"], False)
        observed = HGVSProtChange.fromStr("Ala2[10]")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ala", 2, None, None, None, "dup", [], False)
        observed = HGVSProtChange.fromStr("A2[1]")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ala", 2, None, None, None, "dup", [], False)
        observed = HGVSProtChange.fromStr("Ala2[1]")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ala", 2, "Pro", 5, None, None, ["[", "10", "]"], False)
        observed = HGVSProtChange.fromStr("A2_P5[10]")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ala", 2, "Pro", 5, None, None, ["[", "10", "]"], False)
        observed = HGVSProtChange.fromStr("Ala2_Pro5[10]")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ala", 2, "Pro", 5, None, "dup", [], False)
        observed = HGVSProtChange.fromStr("A2_P5[1]")
        self.assertEqual(observed, expected)
        expected = HGVSProtChange("Ala", 2, "Pro", 5, None, "dup", [], False)
        observed = HGVSProtChange.fromStr("Ala2_Pro5[1]")
        self.assertEqual(observed, expected)


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""Classes and functions for managing and processing complex filters like included filters and combined filters with differents operators."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.4.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

from copy import deepcopy


class FiltersCombiner:
    """
    Wrap as an unique filter a combination of several Filter and FiltersCombiner.

    :Code example:
        .. highlight:: python
        .. code-block:: python

            ...
            age_filter = Filter("<", 10, "age")
            treatment_filter = filter("in", ["20ng", "20ng + pc"], "treatment")
            filters = FiltersCombiner(
                [age_filter, treatment_filter],
                "and",
                "young_with_treatment"
            )
            for curr in patients:
                if filters.eval(curr):
                    print(curr)
    """

    def __init__(self, filters, operator="and", name=None, description=None):
        """
        Build and return an instance of FiltersCombiner.

        :param filters: List of Filter and FiltersCombiner evaluated by the instance.
        :type filters: list
        :param operator: The type of evaluation processed between filters. With "and" all the filters must have an valid evaluation, with "or" at least one filter must have an valid evaluation, with "xor" only one filter must have an valid evaluation.
        :type operator: str
        :param name: The name of the filter.
        :type name: str
        :param description: The description of the filter.
        :type description: str
        :return: The new instance.
        :rtype: filters.FiltersCombiner
        :todo: xor.
        """
        self.description = description
        self.filters = filters
        self.name = name
        self.operator = operator
        if operator not in ["and", "or"]:
            raise Exception(
                'The operator "{}" is not valid for {}.'.format(
                    operator, self.__class__.__name__
                )
            )

    def eval(self, item):
        """
        Return True if the item fits the filters regarding to the operator.

        :param item: The evaluated element.
        :type: *
        :return: True if the item fits the filters regarding to the operator.
        :rtype: bool
        :todo: xor.
        """
        is_valid = True
        nb_filters = len(self.filters)
        if nb_filters > 0:
            is_valid = None
            idx = 0
            while(idx < nb_filters and is_valid is None):
                if self.filters[idx].eval(item):
                    if self.operator == "or":
                        is_valid = True
                else:
                    if self.operator == "and":
                        is_valid = False
                idx += 1
            if is_valid is None:
                if self.operator == "and":  # The operator is AND and no false has been encountered
                    is_valid = True
                else:  # The operator is OR and no true has been encountered
                    is_valid = False
        return is_valid


class Filter:
    """
    Manage one filter.

    :Code example:
        .. highlight:: python
        .. code-block:: python

            patients = [
                {"age": 8, "treatment": "placebo", "group":["A", "B"]},
                {"age": 9, "treatment": "20ng ip", "group":["C"]},
                {"age": 11, "treatment": "placebo", "group":["A", "D"]},
                {"age": 25, "treatment": "20ng ip", "group":["B"]},
                {"age": 30, "treatment": "20ng pc", "group":["C", "B"]},
                {"age": 33, "treatment": "20ng pc", "group":["A", "D"]},
                {"age": 30, "treatment": "placebo", "group":["A", "B"]},
            ]
            # Select children
            children_filter = Filter("<", 13, "age")
            for curr in patients:
                if children_filter.eval(curr):
                    print(curr)
            # Select not placebo
            treatment_filter = Filter("=", "placebo", "treatment", action="exclude")
            for curr in patients:
                if treatment_filter.eval(curr):
                    print(curr)
            # Select patients with at least group A or group C
            group_filter = Filter("in", ["A", "C"], "group", aggregator="nb:1")
            for curr in patients:
                if group_filter.eval(curr):
                    print(curr)
    """

    def __init__(self, operator, values, getter=None, aggregator="ratio:1", name=None, description=None, action="select"):
        """
        Build and return an instance of FiltersCombiner.

        :param values: The value(s) used as threshold/reference in item evaluation.
        :type values: *
        :param operator: The operator used in evaluation.
        :type operator: str
        :param getter: The property or the function applied on evaluated item to get the evaluated value. See Filter.setGetter().
        :type getter: str or callable
        :param aggregator: The rule used when evaluated value is a list. It determines the number of element in list that must fit the filter.
        :type aggregator: str
        :param name: The name of the filter.
        :type name: str
        :param description: The description of the filter.
        :type description: str
        :param action: Determines if the item fitting the filter are kept ("select") or rejected ("exclude") by filter.
        :type action: str
        :return: The new instance.
        :rtype: filters.FiltersCombiner
        """
        self.action = action
        if action not in ["select", "exclude"]:
            raise Exception('The action "{}" is invalid. An action must be "select" or "exclude".'.format(action))
        self.description = description
        self.name = name
        self.values = values
        self.setAggregator(aggregator)
        self.setGetter(getter)
        self.setOperator(operator)

    @staticmethod
    def fromDict(filter_desc):
        """
        Return an instance of Filter corresponding to the parameters in dictionary.

        :param filter_desc: The parameters.
        :type: dict
        :return: An instance of Filter corresponding to the parameters in dictionary.
        :rtype: filters.Filter
        """
        cleaned_desc = deepcopy(filter_desc)
        if "class" in cleaned_desc:
            cleaned_desc.pop('class', None)
        return Filter(**cleaned_desc)

    def toDict(self):
        """
        Return the dictionary corresponding to the instance of Filter.

        :return: The dictionary corresponding to the instance of Filter.
        :rtype: dict
        """
        obj_dict = {"class": "Filter"}
        for attr_name in dir(self):
            if not attr_name.startswith("_"):
                attr_value = self.__getattribute__(attr_name)
                if not callable(attr_value):
                    obj_dict[attr_name] = attr_value
        return obj_dict

    def __eq__(self, other):
        self_dict = {key: val for key, val in self.__dict__.items() if not (key.startswith("_") and key.endswith("Fct"))}
        other_dict = {key: val for key, val in other.__dict__.items() if not (key.startswith("_") and key.endswith("Fct"))}
        return self_dict == other_dict

    def setOperator(self, new):
        """
        Change the operator used in evaluation.

        :param new: The new value of the operator.
        :type new: str
        """
        self.operator = new
        self.setFct()

    def setAggregator(self, new):
        """
        Change the aggregator used to evaluate a list of values.

        :param new: [str] The new value of the aggregator. The aggregator must have one of the following form: "nb:INT" or "ratio:FLOAT". The filter is ok if the evaluated list returned by getter has:
          (1) a number of values fitting the Filter superior than INT if aggregator is nb:INT.
          (2) a ratio of values fitting the Filter superior than FLOAT if aggregator is ratio:FLOAT.
        :type new: str
        """
        self.aggregator = new  # nb:X, ratio:X.X
        agg_type, agg_threshold = new.split(":")
        if agg_type == "nb":
            self._aggregatorEvalFct = lambda nb, length: nb >= int(agg_threshold)
        elif agg_type == "ratio":
            self._aggregatorEvalFct = lambda nb, length: nb >= float(agg_threshold) * length
        else:
            raise AttributeError('The aggregator "{}" is invalid. An aggregator must start with "nb" or "ratio".'.format(new))

    def setGetter(self, new):
        """
        Change the getter used to retrieve evaluated value from evaluated item.

        :param new: [str|callable] The new value of the getter. This value can take 3 forms:
          (1) If the getter is None: the evaluated item himself is used in evaluation.
          (2) If the getter is callable: this is the result of this call applied on item that is evaluated. Use a callable is not compatible with Filter.toDict().
          (3) If the getter is a string: the getter is parsed by getRecordValue to retrieve a specific property.
        :type new: str
        """
        self.getter = new
        self._getterFct = None
        if new is None:
            def returnItem(item):
                return item
            self._getterFct = returnItem
        else:
            if callable(new):
                self._getterFct = new
            else:
                self._getterFct = self.getRecordValue

    def getRecordValue(self, record, _getter=None):
        key = _getter
        if _getter is None:
            key = self.getter
        sub_key = None
        if "." in key:
            key, sub_key = key.split(".", 1)
        # Get value from current key
        value = record
        if key.startswith("m:"):
            method_name = key[2:]
            method_param = None
            if "(" in method_name:  # Parameters must be numeric or string
                if method_name.endswith("()"):
                    method_name = method_name[:-2]
                else:
                    method_name, method_param = method_name.split("(", 1)
                    method_param = [elt.strip() for elt in method_param[:-1].split(",")]
            method = value.__getattribute__(method_name)
            if method_param is None:
                value = method()
            else:
                value = method(*method_param)
        elif key.startswith("i:"):
            if key != "i:":
                value = value.__getattribute__(key[2:])
                if issubclass(value.__class__, dict):
                    new_value = list()
                    for elt_key, elt_value in value.items():
                        new_value.append(elt_value)
                    value = new_value
                elif issubclass(value.__class__, list):
                    new_value = list()
                    for curr_elt in value:
                        new_value.append(curr_elt)
                    value = new_value
                else:
                    raise Exception("The iteration has been ask on non iterable object {}.".format(key[2:]))
        elif issubclass(value.__class__, dict):
            value = value[key]
        elif issubclass(value.__class__, list):
            value = value[int(key)]
        elif issubclass(value.__class__, object):
            value = value.__getattribute__(key)
        # Get value from current sub-key
        if sub_key is not None:
            if key.startswith("i:"):
                new_value = list()
                for curr_elt in value:
                    new_value.append(self.getRecordValue(curr_elt, sub_key))
                value = new_value
            else:
                value = self.getRecordValue(value, sub_key)
        # return
        return value

    def eval(self, item):
        """
        Return True if the item fits the filter and action is select or if the item does not fit the filter and action is exclude.

        :param item: The evaluated element.
        :type item: *
        :return: True if the item fits the filter and action is select or if the item does not fit the filter and action is exclude.
        :rtype: bool
        """
        is_valid = False
        # Eval item
        value = self._getterFct(item)
        if issubclass(value.__class__, list):
            is_valid_list = [self._evalFct(curr_val, self.values) for curr_val in value]
            if self._aggregatorEvalFct(is_valid_list.count(True), len(value)):
                is_valid = True
        else:
            is_valid = self._evalFct(value, self.values)
        # Apply action
        if self.action == "exclude":
            is_valid = not is_valid
        # Return
        return is_valid

    def setFct(self):
        """Set the evaluation function according to the operator."""
        if self.operator in ["=", "==", "eq"]:
            self._evalFct = lambda val, ref: val == ref
        elif self.operator in ["!=", "<>", "ne"]:
            self._evalFct = lambda val, ref: val != ref
        elif self.operator in ["<=", "le"]:
            self._evalFct = lambda val, ref: False if val is None else val <= ref
        elif self.operator in [">=", "ge"]:
            self._evalFct = lambda val, ref: False if val is None else val >= ref
        elif self.operator in ["<", "lt"]:
            self._evalFct = lambda val, ref: False if val is None else val < ref
        elif self.operator in [">", "gt"]:
            self._evalFct = lambda val, ref: False if val is None else val > ref
        elif self.operator == "in":
            self._evalFct = lambda val, ref: val in ref
        elif self.operator == "contains":
            if not issubclass(self.values.__class__, str):
                raise AttributeError('The reference value in filter must be a string for contains operator. If you want to apply "contains" with several possibility you must declare n filters and combine them with FiltersCombiner.')
            self._evalFct = lambda val, ref: ref in val
        elif self.operator == "not in":
            self._evalFct = lambda val, ref: val not in ref
        else:
            raise AttributeError('The operator "{}" is not implemented.'.format(self.operator))


def filtersFromDict(filters):
    """
    Return filter from filters descriptor.

    :param filters: Filters parameters.
    :type filters: dict
    :return: The filter corresponding to descriptor.
    :rtype: filters.Filter or filters.FiltersCombiner
    """
    if "class" not in filters or filters["class"] == "Filter":
        return Filter.fromDict(filters)
    elif filters["class"] == "FiltersCombiner":
        cleaned_desc = deepcopy(filters)
        cleaned_desc.pop('class', None)
        cleaned_desc["filters"] = [filtersFromDict(sub_filter) for sub_filter in filters["filters"]]
        return FiltersCombiner(**cleaned_desc)
    else:
        raise Exception('The class "{}" is invalid for filters dictionary. The value must be "Filter" or "FiltersCombiner".'.format(filters["class"]))



# -*- coding: utf-8 -*-
"""Classes and functions for manipulating/processing regions defines on references sequences."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.8.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'


class Region:
    """Class to manage region on a sequence."""

    def __init__(self, start=None, end=None, strand=None, reference=None, name=None, annot=None):
        """
        Build and return an instance of Region.

        :param start: The start position on the reference. This position is 1-based and ascending (start <= end).
        :type start: int
        :param end: The end position on the reference. This position is 1-based and ascending (start <= end). [Default: start]
        :type end: int
        :param strand: The strand of the instance ("+" or "-").
        :type strand: str
        :param reference: The region object or the region name of the reference.
        :type reference: Region | str
        :param name: The name of the region.
        :type name: str
        :param annot: The annotations of the region.
        :type annot: dict
        :return: The new instance.
        :rtype: region.Region
        """
        if start is not None and end is not None and int(start) > int(end):
            raise Exception("Start must be inferior to end.")
        self.start = None if start is None else int(start)
        self.end = self.start if end is None else int(end)
        self.strand = strand
        self.setReference(reference)
        self.name = name
        self.annot = dict() if annot is None else annot

    def __str__(self):
        """
        Return the “informal” or nicely printable string representation of the object.

        :return: The printable string representation of the object.
        :rtype: str
        """
        clean_str = self.name
        if clean_str is None:
            if self.reference is not None:
                clean_str = self.reference.name
            if self.start is not None:
                clean_str = "{}:{}-{}".format(clean_str, self.start, self.end)
        return clean_str

    def getCoordinatesStr(self):
        """
        Return printable string representation of the coordinates.

        :return: The printable representation of the coordinates.
        :rtype: str
        """
        return "{}:{}-{}[{}]".format(
            (None if self.reference is None else self.reference.name),
            self.start,
            self.end,
            self.strand
        )

    def length(self):
        """
        Return length of the region.

        :return: The length of the region.
        :rtype: int
        """
        return(self.end - self.start + 1)

    def setReference(self, reference):
        """
        Change the reference region of the instance.

        :param reference: The region object or the region name of the reference.
        :type regreferenceion_pos: region.Region | str
        """
        if reference is None:
            self.reference = None
        elif isinstance(reference, str):
            self.reference = Region(name=reference)
        elif isinstance(reference, Region):
            self.reference = reference
        else:
            raise Exception("The type " + str(type(reference)) + " is not authorized for reference attribute in Region class.")

    def getPosOnRef(self, region_pos):
        """
        Return coordinate on reference sequence from the coordinate on self.

        :param region_pos: Coordinate on region (1-based).
        :type region_pos: int.
        :return: The coordinate on reference sequence (1-based).
        :rtype: int
        """
        if self.strand is None:
            raise Exception("Cannot return a reference position from the region position because the strand is None in {}.".format(
                self
            ))
        reference_pos = None
        if self.strand == '+':
            reference_pos = region_pos + self.start - 1
        else:
            reference_pos = self.end - region_pos + 1
        return reference_pos

    def getPosOnRegion(self, ref_pos):
        """
        Return coordinate on region from the coordinate on reference sequence.

        :param ref_pos: The coordinate on reference sequence (1-based).
        :type ref_pos: int
        :return: Coordinate on region (1-based).
        :rtype: int
        """
        if self.strand is None:
            raise Exception("Cannot return a region position from the reference position because the strand is None in {}.".format(
                self.name
            ))
        if ref_pos < self.start or ref_pos > self.end:
            raise ValueError("The region {} does not contains the position {}.".format(
                self, ref_pos
            ))
        region_pos = None
        if self.strand == '+':
            region_pos = ref_pos - self.start + 1
        else:
            region_pos = self.end - ref_pos + 1
        return region_pos

    def contains(self, eval_region):
        """
        Return True if the region contains the eval_region.

        :param eval_region: The evaluated region.
        :type eval_region: Region
        :return: True if the region contains the evaluated region.
        :rtype: bool
        """
        contains = False
        if self.reference.name == eval_region.reference.name:
            if self.start <= eval_region.start and self.end >= eval_region.end:
                contains = True
        return contains

    def strandedContains(self, eval_region):
        """
        Return True if the region contains the eval_region and their are on the same strand.

        :param eval_region: The evaluated region.
        :type eval_region: Region
        :return: True if the region contains the evaluated region.
        :rtype: bool
        """
        contains = False
        if self.strand == eval_region.strand:
            contains = self.contains(eval_region)
        return contains

    def hasOverlap(self, eval_region):
        """
        Return True if the region has an overlap with eval_region.

        :param eval_region: The evaluated region.
        :type eval_region: Region
        :return: True if the region has an overlap with evaluated region.
        :rtype: bool
        """
        has_overlap = False
        if self.reference.name == eval_region.reference.name:
            if not self.start > eval_region.end and not self.end < eval_region.start:
                has_overlap = True
        return has_overlap

    def hasStrandedOverlap(self, eval_region):
        """
        Return True if the region has an overlap with eval_region and their are on the same strand.

        :param eval_region: The evaluated region.
        :type eval_region: Region
        :return: True if the region has an overlap with evaluated region.
        :rtype: bool
        """
        has_overlap = False
        if self.strand == eval_region.strand:
            has_overlap = self.hasOverlap(eval_region)
        return has_overlap

    def getMinDist(self, eval_region):
        """
        Return the distance with the eval_region.

        :param eval_region: The evaluated region.
        :type eval_region: Region
        :return: The distance between the instance and the evaluated region.
        :rtype: int
        """
        if self.reference.name != eval_region.reference.name:
            raise Exception('The minimal distance between regions cannot be processed because their are located on diffrents reference ("' + self.reference.name + '" vs "' + eval_region.reference.name + '").')
        min_dist = None
        if self.hasOverlap(eval_region):  # Eval region overlaps region
            min_dist = 0
        elif eval_region.end < self.start:  # Eval region is located before region
            min_dist = self.start - eval_region.end
        else:  # Eval region is located after region
            min_dist = eval_region.start - self.end
        return min_dist

    @staticmethod
    def fromStr(interval_str):
        """
        Return region instance from interval string (example: 5:15746211-5746311).

        :param interval_str: interval string in format <ref_name>:<start_pos>-<end_pos>. Positions are 1-based.
        :type interval_str: str
        :return: Region instance from interval string.
        :rtype: anacore.region.Region
        """
        reference, start_end = interval_str.split(":")
        start, end = start_end.split("-")
        return Region(start, end, None, reference)


class RegionTree(Region):
    """Class to manage region with hierarchical relations (example: gene -> transcript -> exons)."""

    def __init__(self, start=None, end=None, strand=None, reference=None, name=None, annot=None, parent=None, children=None):
        """
        Build and return an instance of RegionTree.

        :param start: The start position on the reference. This position is 1-based and ascending (start <= end).
        :type start: int
        :param end: The end position on the reference. This position is 1-based and ascending (start <= end). [Default: start]
        :type end: int
        :param strand: The strand of the instance ("+" or "-").
        :type strand: str
        :param reference: The region object or the region name of the reference.
        :type reference: region.Region | str
        :param name: The name of the region.
        :type name: str
        :param annot: The annotations of the region.
        :type annot: dict
        :param parent: The parent region (example: gene for a transcript).
        :type parent: region.RegionTree
        :param children: The list of sub-regions (example: exons in transcript).
        :type children: region.RegionList
        :return: The new instance.
        :rtype: region.RegionTree
        """
        self.children = RegionList()
        Region.__init__(self, start, end, strand, reference, name, annot)
        if children is not None:
            for child in children:
                self.addChild(child)
        self.parent = parent
        if parent is not None:
            parent.addChild(self)

    def __str__(self):
        """
        Return the “informal” or nicely printable string representation of the object.

        :return: The printable string representation of the object.
        :rtype: str
        """
        clean_str = Region.__str__(self)
        clean_str = "{} [{} children]".format(clean_str, len(self.children))
        if "feature" in self.annot and self.annot["feature"] is not None:
            clean_str = "{} {}".format(self.annot["feature"], clean_str)
        return clean_str

    def __getattribute__(self, name):
        """
        Return the value of the selected attribute.

        :param name: The name of the selected attribute.
        :type name: str
        :return: The value of the attribute.
        :rtype: *
        """
        value = object.__getattribute__(self, name)
        if value is None:
            if name == "start":
                if len(self.children) > 0:
                    value = min([child.start for child in self.children])
            elif name == "end":
                if len(self.children) > 0:
                    value = max([child.end for child in self.children])
            elif name == "strand":
                if len(self.children) > 0:
                    value = self.children[0].strand
            elif name == "reference":
                if len(self.children) > 0:
                    value = self.children[0].reference
        return value

    def addChild(self, child):
        """
        Add child region in region.

        :param child: The added region.
        :type child: region.RegionTree
        """
        # Check compatibility
        if self.strand is not None and child.strand is not None:
            if self.strand != child.strand:
                raise ValueError("The sub-region {} cannot be added to {} because their strands are different.".format(child, self))
        if self.reference is not None:
            if self.reference.name != child.reference.name:
                raise ValueError("The sub-region {} cannot be added to {} because their reference region are different.".format(child, self))
        # Process
        child.parent = self
        self.children.insert(0, child)
        if child.start is not None:
            self.sortChildren()

    def sortChildren(self):
        """Sort children in order of their apparition on self strand. Add siblings_idx (1-based) in annotations of children."""
        if self.strand is None:
            raise Exception("Cannot sort sub-regions because the strand is None in {}.".format(self))
        # Sort
        if self.strand == "-":
            self.children = RegionList(sorted(self.children, key=lambda x: (x.end, x.start), reverse=True))
        else:
            self.children = RegionList(sorted(self.children, key=lambda x: (x.start, x.end)))
        # Add siblings_idx
        for siblings_idx, child in enumerate(self.children):
            child.annot["siblings_idx"] = siblings_idx + 1  # Siblings idx is 1-based


class RegionList(list):
    """Class to manage regions list. for example all the gene on a chromosome."""

    def __init__(self, regions=None):
        """
        Build and return an instance of RegionList.

        :param regions: List of regions.
        :type regions: list
        :return: The new instance.
        :rtype: region.RegionList
        """
        if regions is not None:
            for curr_region in regions:
                self.append(curr_region)

    def getContainers(self, eval_region):
        """
        Return all the regions that contains the eval_region.

        :param eval_region: The evaluated region.
        :type eval_region: Region
        :return: Regions containing the evaluated region.
        :rtype: list
        """
        containers = list()
        for curr_region in self:
            if curr_region.contains(eval_region):
                containers.append(curr_region)
        return containers

    def getOverlapped(self, eval_region):
        """
        Return all the regions that have an overlap with eval_region.

        :param eval_region: The evaluated region.
        :type eval_region: Region
        :return: Regions having an overlap with evaluated region.
        :rtype: list
        """
        overlapped = list()
        for curr_region in self:
            if curr_region.hasOverlap(eval_region):
                overlapped.append(curr_region)
        return overlapped

    def getNearests(self, eval_region, select_fct=None):
        """
        Return the nearest region to eval_region.

        :param eval_region: The evaluated region.
        :type eval_region: Region
        :return: The distance between the nearest region and self.
        :rtype: list
        """
        nearests = list()
        min_dist = None
        for curr_region in self:
            selected = True
            if select_fct is not None:
                selected = select_fct(curr_region)
            if selected:
                if curr_region.reference.name == eval_region.reference.name:
                    curr_dist = curr_region.getMinDist(eval_region)
                    if min_dist is None or curr_dist <= min_dist:
                        if min_dist is None or curr_dist < min_dist:
                            nearests = list()
                        min_dist = curr_dist
                        nearests.append(curr_region)
        return min_dist, nearests


def splittedByRef(region_list):
    """
    Return regions list by reference.

    :param region_list: The list of regions.
    :type region_list: RegionList
    :return: Regions list by reference.
    :rtype: dict
    """
    regions_by_ref = {}
    for curr_region in region_list:
        ref_name = curr_region.reference.name
        if ref_name not in regions_by_ref:
            regions_by_ref[ref_name] = RegionList()
        regions_by_ref[ref_name].append(curr_region)
    return regions_by_ref


def mergedRegion(regions, trace=False):
    """
    Return a region corresponding to the merge of all provided regions.

    :param regions: One list of overlapped regions.
    :param type: list
    :param trace: if true the provided regions are store in merged_region.annot["merge_traceback"].
    :type trace: bool
    :return: The region corresponding to the merge of all provided regions.
    :rtype: Region
    """
    # Check reference
    ref = set()
    for curr_region in regions:
        if curr_region.reference is None:
            ref.add(None)
        else:
            ref.add(curr_region.reference.name)
    if len(ref) > 1:
        raise Exception("All the regions in mergedRegion are not defined on the same reference.")
    # Init merged region
    merged_region = Region(
        min([curr.start for curr in regions]),
        max([curr.end for curr in regions]),
        None,
        regions[0].reference,
        None
    )
    # Strand
    strands = list(set([curr.strand for curr in regions]))
    if len(strands) == 1 and strands[0] is not None:
        merged_region.strand = strands[0]
    # Traceback
    if trace:
        merged_region.annot["merge_traceback"] = RegionList()
        for curr_region in regions:
            if "merge_traceback" in curr_region.annot:
                for curr_trace in curr_region.annot["merge_traceback"]:
                    merged_region.annot["merge_traceback"].append(curr_trace)
            else:
                merged_region.annot["merge_traceback"].append(curr_region)
    return merged_region


def consolidated(regions, merge_contiguous=True, trace=False):
    """
    Return consolidated regions: the overlapping regions (and optionally the contiguous) are merged in one region.

    :param regions: List of regions to consolidate.
    :param type: list
    :param merge_contiguous: If true the contiguous regions are also merged.
    :type merge_contiguous: bool
    :param trace: if true the provided regions are store in merged_region.annot["merge_traceback"].
    :type trace: bool
    :return: Consolidated regions.
    :rtype: RegionList
    """
    padding = 1 if merge_contiguous else 0
    merged_regions = RegionList()
    regions_by_ref = splittedByRef(regions)
    for ref_id, ref_regions in sorted(regions_by_ref.items()):
        sorted_regions = sorted(ref_regions, key=lambda x: (x.start, x.end))
        prev_region = sorted_regions[0]
        for curr_region in sorted_regions[1:]:
            if curr_region.start <= prev_region.end + padding:
                prev_region = mergedRegion([prev_region, curr_region], trace)
            else:
                merged_regions.append(prev_region)
                prev_region = curr_region
        merged_regions.append(prev_region)
    return merged_regions


def iterOverlapped(queries, subjects, check_ref=True):
    """
    Return a generator on query and his overlapped subjects. For example you can provide a list of variant (queries) and search the genes overlapped by each variant in the list of genes (subjects).

    :param queries: The queries. They must be defined on the same reference.
    :type queries: list
    :param subjects: The regions where the overlap is searched. They must be defined on the same reference.
    :type subjects: list
    :param check_ref: If True all the queries and subjects are checked to ensure they care defined on the same reference. Use False to gain computation time if you are certain that all elements are defined on the same reference.
    :type check_ref: bool
    :return: Each iteration return one query and the list of subjects overlapped by it. The queries are ordered by ascending coordinates.
    :rtype: A generator on couple (Region, RegionList)
    """
    if check_ref:
        references = set()
        for elt in subjects + queries:
            if elt.reference is None:
                references.add(None)
            else:
                references.add(elt.reference.name)
        if len(references) > 1:
            raise Exception("All the queries and subjects in iterOverlapped are not defined on the same reference.")
    subjects = sorted(subjects, key=lambda x: (x.start, x.end))
    queries = sorted(queries, key=lambda x: (x.start, x.end))
    subject_idx = 0
    nb_subjects = len(subjects)
    for curr_query in queries:
        # Find overlapping transcripts
        while subject_idx < nb_subjects and curr_query.start > subjects[subject_idx].end:
            subject_idx += 1
        first_subject_idx = None
        overlapping_subjects = RegionList()
        while subject_idx < nb_subjects and curr_query.end >= subjects[subject_idx].start:
            curr_subject = subjects[subject_idx]
            if not curr_query.start > curr_subject.end:
                overlapping_subjects.append(curr_subject)
                if first_subject_idx is None:
                    first_subject_idx = subject_idx
            subject_idx += 1
        subject_idx = (subject_idx if first_subject_idx is None else first_subject_idx)  # Return to the first subject overlapping the query
        # Return current query and his overlaps
        yield(curr_query, overlapping_subjects)


def iterOverlappedByRegion(queries_by_chr, subject_by_chr):
    """
    Return a generator on query and his overlapped subjects. For example you can provide a list of variant (queries) and search the genes overlapped by each variant in the list of genes (subjects).

    :param queries_by_chr: By chromosome the queries.
    :type queries_by_chr: dict
    :param subject_by_chr: By chromosome the regions where the overlap is searched.
    :type subject_by_chr: dict
    :return: Each iteration return the chromosome name, one query and the list of subjects overlapped by it. The queries are ordered by ascending coordinates.
    :rtype: A generator on couple (str, Region, RegionList)
    """
    for chrom in sorted(queries_by_chr):
        if chrom not in subject_by_chr:
            for query in queries_by_chr[chrom]:
                yield(chrom, query, RegionList())
        else:
            for query, overlapping_subjects in iterOverlapped(queries_by_chr[chrom], subject_by_chr[chrom], False):
                yield(chrom, query, overlapping_subjects)



#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2021 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.0.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import sys
import tempfile
import unittest
import uuid

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.bed import BEDIO, BEDRecord, getSortedAreasByChr
from anacore.region import Region


########################################################################
#
# FUNCTIONS
#
########################################################################
def diffRecord(record_a, record_b):
    are_diff = False
    attr_list = [
        "start", "end", "chrom", "score", "thickStart", "thickEnd",
        "itemRgb", "blockCount", "blockSizes", "blockStarts"
    ]
    for attr_name in attr_list:
        if str(getattr(record_a, attr_name)) != str(getattr(record_b, attr_name)):
            are_diff = True
    return are_diff


class TestBEDRecord(unittest.TestCase):
    def testSetAttrChrom(self):
        record = BEDRecord("7", 127471197, 127472363)
        record.chrom = "chr7"
        self.assertEqual(record.chrom, "chr7")
        self.assertEqual(record.reference.name, "chr7")


class TestBEDIO(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_in = os.path.join(tmp_folder, unique_id + "_in.bed")
        self.tmp_out = os.path.join(tmp_folder, unique_id + "_out.bed")

        self.expected_records = [
            BEDRecord(
                "chr7", 127471197, 127472363, "Pos1", 0, "+",
                127471197, 127472363, [255, 0, 0], None, None, None
            ),
            BEDRecord(
                "chr1", 137475865, 137477031, "Pos2", 0, "+",
                137475865, 137477031, None, None, None, None
            ),
            BEDRecord(
                "chr1", 127475865, 127477031, "Neg1", 0, "-",
                127475865, 127477031, [0, 0, 255], None, None, None
            ),
            BEDRecord(
                "chr22", 1001, 5000, "cloneA", 960, "+",
                1001, 5000, None, 2, [567, 488], [0, 3512]
            ),
            BEDRecord(
                "chr22", 2001, 6000, "cloneB", 900, "-",
                2001, 6000, None, 2, [433, 399], [0, 3601]
            )
        ]
        self.expected_content = """browser hide all
track name="ItemRGBDemo" description="Item RGB demonstration" visibility=2 itemRgb="On"
chr7	127471196	127472363	Pos1	0	+	127471196	127472363	255,0,0
chr1	137475864	137477031	Pos2	0	+	137475864	137477031
chr1	127475864	127477031	Neg1	0	-	127475864	127477031	0,0,255
track name=pairedReads description="Clone Paired Reads" useScore=1
chr22	1000	5000	cloneA	960	+	1000	5000	0	2	567,488	0,3512
chr22	2000	6000	cloneB	900	-	2000	6000	0	2	433,399	0,3601"""
        with open(self.tmp_in, "w") as writer:
            writer.write(self.expected_content)

    def testGetMaxNbCol(self):
        self.assertEqual(
            BEDIO.getMaxNbCol(self.tmp_in),
            12
        )

    def testGetSortedAreasByChr(self):
        # Read
        observed = getSortedAreasByChr(self.tmp_in)
        # Expected
        expected = {
            "chr1": [
                BEDRecord(
                    "chr1", 127475865, 127477031, "Neg1", 0, "-",
                    127475865, 127477031, [0, 0, 255]
                ),
                BEDRecord(
                    "chr1", 137475865, 137477031, "Pos2", 0, "+",
                    137475865, 137477031, None
                )
            ],
            "chr7": [
                BEDRecord(
                    "chr7", 127471197, 127472363, "Pos1", 0, "+",
                    127471197, 127472363, [255, 0, 0]
                )
            ],
            "chr22": [
                BEDRecord(
                    "chr22", 1001, 5000, "cloneA", 960, "+",
                    1001, 5000, None, 2, [567, 488], [0, 3512]
                ),
                BEDRecord(
                    "chr22", 2001, 6000, "cloneB", 900, "-",
                    2001, 6000, None, 2, [433, 399], [0, 3601]
                )
            ]
        }
        # Assert
        self.assertTrue(len(expected) == len(observed))
        for chrom in expected:
            for record_a, record_b in zip(expected[chrom], observed[chrom]):
                self.assertFalse(diffRecord(record_a, record_b))

    def testIsValid(self):
        # Valid
        self.assertTrue(
            BEDIO.isValid(self.tmp_in)
        )
        # Invalid pos type
        with open(self.tmp_out, "w") as writer:
            writer.write(self.expected_content)
            writer.write("\nchr1	137475864	test	Neg2	0	-")
        self.assertFalse(
            BEDIO.isValid(self.tmp_out)
        )
        # Invalid strand value
        with open(self.tmp_out, "w") as writer:
            writer.write(self.expected_content)
            writer.write("\nchr1	137475864	137475868	Neg2	0	0")
        self.assertFalse(
            BEDIO.isValid(self.tmp_out)
        )
        # Empty RGB value
        with open(self.tmp_out, "w") as writer:
            writer.write(self.expected_content)
            writer.write("\nchr7	127471196	127472363	testRGBNone	0	+	127471196	127472363	.")
        self.assertTrue(
            BEDIO.isValid(self.tmp_out)
        )
        # Invalid RGB value
        with open(self.tmp_out, "w") as writer:
            writer.write(self.expected_content)
            writer.write("\nchr7	127471196	127472363	Pos1	0	+	127471196	127472363	255")
        self.assertFalse(
            BEDIO.isValid(self.tmp_out)
        )

    def testRead(self):
        # Read
        observed_records = []
        with BEDIO(self.tmp_in) as reader:
            for record in reader:
                observed_records.append(record)
        # Assert
        self.assertTrue(len(self.expected_records) == len(observed_records))
        for record_a, record_b in zip(self.expected_records, observed_records):
            self.assertFalse(diffRecord(record_a, record_b))

    def testWriteFromBEDRecord(self):
        expected = """chr7	127471196	127472363	Pos1	0	+	127471196	127472363	255,0,0
chr1	137475864	137477031	Pos2	0	+	137475864	137477031	.
chr1	127475864	127477031	Neg1	0	-	127475864	127477031	0,0,255"""
        with BEDIO(self.tmp_out, "w", 9) as writer:
            for curr_region in self.expected_records[:3]:
                writer.write(curr_region)
        observed = None
        with open(self.tmp_out) as reader:
            observed = "".join(reader.readlines()).strip()
        self.assertEqual(observed, expected)

    def testWriteFromRegions(self):
        regions = [
            Region(158, 367, "-", "chr1", "target_01"),
            Region(54899, 54999, "+", "chr1", "target_02"),
        ]

        # From regions and over columns
        expected = """chr1\t157\t367\ttarget_01\t.\t-\t.\t.\t.
chr1\t54898\t54999\ttarget_02\t.\t+\t.\t.\t."""
        with BEDIO(self.tmp_out, "w", 9) as writer:
            for curr_region in regions:
                writer.write(curr_region)
        observed = None
        with open(self.tmp_out) as reader:
            observed = "".join(reader.readlines()).strip()
        self.assertEqual(observed, expected)

        # From regions and limit columns
        expected = """chr1\t157\t367
chr1\t54898\t54999"""
        with BEDIO(self.tmp_out, "w", 3) as writer:
            for curr_region in regions:
                writer.write(curr_region)
        observed = None
        with open(self.tmp_out) as reader:
            observed = "".join(reader.readlines()).strip()
        self.assertEqual(observed, expected)

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_in, self.tmp_out]:
            if os.path.exists(curr_file):
                os.remove(curr_file)


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.1.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import sys
import warnings
import unittest

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.filters import Filter, FiltersCombiner, filtersFromDict


########################################################################
#
# FUNCTIONS
#
########################################################################
class TestFilter(unittest.TestCase):
    def setUp(self):
        # Data
        self.data = [
            {"age": 8, "treatment": "placebo", "group":["A", "B"]},
            {"age": 9, "treatment": "20ng ip", "group":["C"]},
            {"age": 11, "treatment": "placebo", "group":["A", "D"]},
            {"age": 25, "treatment": "20ng ip", "group":["B"]},
            {"age": 30, "treatment": "20ng pc", "group":["C", "B"]},
            {"age": 33, "treatment": "20ng pc", "group":["A", "D"]},
            {"age": 30, "treatment": "placebo", "group":["A", "B"]},
        ]


    def testFromDict(self):
        # Only required attributes are provided
        filter_obj = Filter.fromDict({
            "class": "Filter",
            "operator": "<",
            "values": 25
        })
        expected = Filter("<", 25)
        self.assertEqual(expected, filter_obj)
        # All attributes are provided
        filter_obj = Filter.fromDict({
            "action": "exclude",
            "aggregator": "nb:1",
            "class": "Filter",
            "description": "Select patients with age lower than 25 years",
            "getter": "age",
            "name": "young",
            "operator": "<",
            "values": 25
        })
        expected = Filter("<", 25, "age", "nb:1", "young", "Select patients with age lower than 25 years", "exclude")
        self.assertEqual(expected, filter_obj)


    def testToDict(self):
        # Only required attributes are provided
        filter_obj = Filter("<", 25)
        expected = {
            "action": "select",
            "aggregator": "ratio:1",
            "class": "Filter",
            "description": None,
            "getter": None,
            "name": None,
            "operator": "<",
            "values": 25
        }
        self.assertEqual(expected, filter_obj.toDict())
        # All attributes are provided
        filter_obj = Filter("<", 25, "age", "nb:1", "young", "Select patients with age lower than 25 years", "exclude")
        expected = {
            "action": "exclude",
            "aggregator": "nb:1",
            "class": "Filter",
            "description": "Select patients with age lower than 25 years",
            "getter": "age",
            "name": "young",
            "operator": "<",
            "values": 25
        }
        self.assertEqual(expected, filter_obj.toDict())


    def testSetFct(self):
        # Test operator contains with list threshold
        with self.assertRaises(AttributeError) as context:
            Filter("contains", ["ip", "pc"], "treatment")  # Select patients with ip or pc in treatment text

        # Test unknown operator
        with self.assertRaises(AttributeError) as context:
            Filter("unknown", ["ip", "pc"], "treatment")


    def testEval(self):
        # Test action: select AND numeric threshold
        filter_obj = Filter("<", 13, "age")  # Select patients with age not < 13
        expected = [True, True, True, False, False, False, False]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test action: exclude AND numeric threshold
        filter_obj = Filter("<", 25, "age", action="exclude")  # Select patients with age not < 25 (>=25)
        expected = [False, False, False, True, True, True, True]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test action: select AND string threshold
        filter_obj = Filter("!=", "placebo", "treatment")  # Select patients with treatment different of placebo
        expected = [False, True, False, True, True, True, False]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test action: exclude AND string threshold
        filter_obj = Filter("=", "20ng pc", "treatment", action="exclude")  # Select patients with treatment different of 20ng pc
        expected = [True, True, True, True, False, False, True]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test operator: contains AND string threshold
        filter_obj = Filter("contains", "20ng", "treatment")  # Select patients with 20ng in treatment text
        expected = [False, True, False, True, True, True, False]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test operator: in AND list threshold
        filter_obj = Filter("in", ["20ng pc", "placebo"], "treatment")  # Select patients with "20ng pc" or "placebo" as treatment
        expected = [True, False, True, False, True, True, True]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test aggregator: nb:1
        filter_obj = Filter("=", "A", "group", aggregator="nb:1")  # Select patients with at least the group A
        expected = [True, False, True, False, False, True, True]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test aggregator: ratio:1
        filter_obj = Filter("=", "C", "group", aggregator="ratio:1")  # Select patients with 100% of grooups are equals to C
        expected = [False, True, False, False, False, False, False]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test aggregator: ratio:0.5
        filter_obj = Filter("=", "C", "group", aggregator="ratio:0.5")  # Select patients with 50% of grooups are equals to C
        expected = [False, True, False, False, True, False, False]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test list threshold AND operator: in AND aggregator: nb:1
        filter_obj = Filter("in", ["A", "C"], "group", aggregator="nb:1")  # Select patients with at least A or C in groups
        expected = [True, True, True, False, True, True, True]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test list threshold AND operator: not in AND aggregator: nb:1
        filter_obj = Filter("not in", ["B", "C"], "group", aggregator="nb:1")  # Select patients with at least one group not in B and C
        expected = [True, False, True, False, False, True, True]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test list threshold AND aggregator: ratio:1
        filter_obj = Filter("in", ["B", "C"], "group", aggregator="ratio:1")  # Select patients where groups are B and/or C
        expected = [False, True, False, True, True, False, False]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        filter_obj = Filter("not in", ["B", "C"], "group", aggregator="ratio:1")  # Select patients where neither group is B or C
        expected = [False, False, True, False, False, True, False]
        observed = [filter_obj.eval(curr) for curr in self.data]
        self.assertEqual(expected, observed)

        # Test None behavior in simple filter
        data_age = [{"age": None}, {"age": 10}, {"age": 1}]
        filter_obj = Filter("<", 5, "age")
        expected = [False, False, True]
        observed = [filter_obj.eval(curr) for curr in data_age]
        self.assertEqual(expected, observed)
        filter_obj = Filter(">", 5, "age")
        expected = [False, True, False]
        observed = [filter_obj.eval(curr) for curr in data_age]
        self.assertEqual(expected, observed)
        filter_obj = Filter("==", None, "age")
        expected = [True, False, False]
        observed = [filter_obj.eval(curr) for curr in data_age]
        self.assertEqual(expected, observed)
        filter_obj = Filter("<>", None, "age")
        expected = [False, True, True]
        observed = [filter_obj.eval(curr) for curr in data_age]
        self.assertEqual(expected, observed)
        filter_obj = Filter("in", [10, 1], "age")
        expected = [False, True, True]
        observed = [filter_obj.eval(curr) for curr in data_age]
        self.assertEqual(expected, observed)
        filter_obj = Filter("not in", [10, 1], "age")
        expected = [True, False, False]
        observed = [filter_obj.eval(curr) for curr in data_age]
        self.assertEqual(expected, observed)

        # Test None behavior in aggregated
        data_class = [{"class": None}, {"class": [None, "A"]}, {"class": ["B"]}]
        filter_obj = Filter("==", "A", "class", "nb:1")
        expected = [False, True, False]
        observed = [filter_obj.eval(curr) for curr in data_class]
        self.assertEqual(expected, observed)
        filter_obj = Filter("<>", "A", "class", "ratio:1")
        expected = [True, False, True]
        observed = [filter_obj.eval(curr) for curr in data_class]
        self.assertEqual(expected, observed)
        # filter_obj = Filter("==", None, "class", "nb:1")
        # expected = [False, True, False]
        # observed = [filter_obj.eval(curr) for curr in data_class]
        # self.assertEqual(expected, observed)
        filter_obj = Filter("not in", ["A", "B"], "class", "nb:1")
        expected = [True, True, False]
        observed = [filter_obj.eval(curr) for curr in data_class]
        self.assertEqual(expected, observed)
        filter_obj = Filter("not in", ["A", "B"], "class", "ratio:1")
        expected = [True, False, False]
        observed = [filter_obj.eval(curr) for curr in data_class]
        self.assertEqual(expected, observed)
        filter_obj = Filter("in", [None, "B"], "class", "ratio:1")
        expected = [True, False, True]
        observed = [filter_obj.eval(curr) for curr in data_class]
        self.assertEqual(expected, observed)
        filter_obj = Filter("in", [None, "B"], "class", "nb:1")
        expected = [True, True, True]
        observed = [filter_obj.eval(curr) for curr in data_class]
        self.assertEqual(expected, observed)


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    warnings.warn("TODO: TestFiltersCombiner")
    warnings.warn("TODO: testFiltersFromDict")
    warnings.warn("TODO: TestFilter.testGetRecordValue")
    unittest.main()



# -*- coding: utf-8 -*-
"""Classes and functions for reading/writing picard tools outputs."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2019 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.0.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'


def castCol(values, col_type):
    """
    Cast all the values to the column type.

    :param values: Values in the column.
    :type val: list
    :param col_type: Column type among "float" or "int" or None or "str".
    :type col_type: str
    """
    cast_fct = eval(col_type)
    for curr_idx, curr_val in enumerate(values):
        if curr_val == "":
            values[curr_idx] = None
        else:
            values[curr_idx] = cast_fct(curr_val)


def getColType(values):
    """
    Return the column type from these values.

    :param values: Values in the column.
    :type val: list
    :return: Type of the value among "float" or "int" or None or "str".
    :rtype: str
    """
    types = {getValType(curr_val) for curr_val in values if curr_val != "" and curr_val is not None}
    col_type = "int" if len(types) != 0 else None
    if "str" in types:
        col_type = "str"
    elif "float" in types:
        col_type = "float"
    return col_type


def getValType(val):
    """
    Return type of the value.

    :param val: Evaluated value.
    :type val: *
    :return: Type of the value among "float" or "int" or None or "str".
    :rtype: str
    """
    val_type = None
    if val is not None:
        val_type = "str"
        if val.isdigit():
            val_type = "int"
        elif val.replace(".", "", 1).isdigit():
            val_type = "float"
    return val_type


class PicardReader(object):
    """Reader for picard tools outputs."""

    def __init__(self, filepath):
        """
        Build and return an instance of PicardReader.

        :param filepath: The filepath.
        :type filepath: str
        :return: The new instance.
        :rtype: anacore.picardIO.PicardReader
        """
        self.filepath = filepath
        self.command = None
        self.metrics = None
        self.histogram = None
        self._parse()

    def _parse(self):
        """
        Read file and store his information.

        :param FH_in: File handle opens to the first line of the picard's file.
        :type FH_in: TextIOWrapper
        """
        with open(self.filepath) as FH_in:
            for curr_line in FH_in:
                if curr_line.startswith("## METRICS CLASS"):
                    self._parseMetrics(FH_in)
                elif curr_line.startswith("## HISTOGRAM"):
                    self._parseHistogram(FH_in)
                elif curr_line.startswith("## "):
                    self._parseHeader(FH_in)

    def _parseHeader(self, FH_in):
        """
        Read header in file and store his information in self.command.

        :param FH_in: File handle opens to the first line of the picard's file.
        :type FH_in: TextIOWrapper
        """
        self.command = FH_in.readline()[1:].strip()
        curr_line = FH_in.readline()
        while curr_line is not None and curr_line.strip() != "":  # Same section
            curr_line = FH_in.readline()

    def _parseHistogram(self, FH_in):
        """
        Read histogram section in file and store his information in self.histogram.

        :param FH_in: File handle opens to the histogram section of the picard's file: after tag "## HISTOGRAM".
        :type FH_in: TextIOWrapper
        """
        self.histogram = dict()
        # Parse header
        header_titles = [elt.strip() for elt in FH_in.readline().split("\t")]
        titles_by_idx = {}
        for curr_idx, curr_title in enumerate(header_titles):
            titles_by_idx[curr_idx] = curr_title
            self.histogram[curr_title] = list()
        # Get values
        curr_line = FH_in.readline()
        while curr_line is not None and curr_line.strip() != "":  # Same section
            for curr_idx, curr_val in enumerate(curr_line.split("\t")):
                col_name = titles_by_idx[curr_idx]
                col_val = curr_val.strip()
                self.histogram[col_name].append(col_val)
            curr_line = FH_in.readline()
        # Apply types
        for curr_title in header_titles:
            col_type = getColType(self.histogram[curr_title])
            if col_type != "str" and col_type is not None:
                castCol(self.histogram[curr_title], col_type)

    def _parseMetrics(self, FH_in):
        """
        Read metrics section in file and store his information in self.metrics.

        :param FH_in: File handle opens to the metrics section of the picard's file: after tag "## METRICS CLASS".
        :type FH_in: TextIOWrapper
        """
        self.metrics = list()
        # Parse header
        header_titles = [elt.strip() for elt in FH_in.readline().split("\t")]
        titles_by_idx = {curr_idx: curr_title for curr_idx, curr_title in enumerate(header_titles)}
        # Get values
        curr_line = FH_in.readline()
        while curr_line is not None and curr_line.strip() != "":  # Same section
            lib_metrics = {curr_title: None for curr_title in header_titles}
            for curr_idx, curr_val in enumerate(curr_line.split("\t")):
                col_name = titles_by_idx[curr_idx]
                col_val = curr_val.strip()
                lib_metrics[col_name] = col_val
            self.metrics.append(lib_metrics)
            curr_line = FH_in.readline()
        # Apply types
        for curr_title in header_titles:
            col_type = getColType([elt[curr_title] for elt in self.metrics])
            if col_type != "str" and col_type is not None:
                cast_fct = eval(col_type)
                for library in self.metrics:
                    if library[curr_title] == "":
                        library[curr_title] = None
                    else:
                        library[curr_title] = cast_fct(library[curr_title])



# -*- coding: utf-8 -*-
"""Classes and functions for reading/writing TopHatFusion outputs."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.1.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

from anacore.abstractFile import AbstractFile
from anacore.region import Region


class TopHatFusionIO(AbstractFile):
    """Class to manage read and write in TopHatFusionIO file (see: https://ccb.jhu.edu/software/tophat/fusion_manual.shtml)."""

    def _parseLine(self):
        """
        Return a structured record from the TopHatFusionIO current line.

        :return: The record.
        :rtype: dict
        """
        fusion, trash_1, contig_a, contig_b, depth_a, depth_b, mate_distances = [elt.strip() for elt in self.current_line.split('@')]
        chrom, break_a, break_b, orientation, nb_splitted_reads, nb_splitted_pairs, nb_pairs_splitted_reads, nb_contradict, base_cover_left, base_cover_right, trash_1 = [field.strip() for field in fusion.split("\t")]
        chrom_a, chrom_b = chrom.split("-")
        break_a = int(break_a)
        break_b = int(break_b)
        strand_a, strand_b = [("+" if elt == "f" else "-") for elt in orientation]
        return {
            "partner_a": Region(break_a, break_a, strand_a, chrom_a),
            "partner_b": Region(break_b, break_b, strand_b, chrom_b),
            "nb_splitted_reads": int(nb_splitted_reads),
            "nb_splitted_pairs": int(nb_splitted_pairs),
            "nb_pairs_splitted_reads": int(nb_pairs_splitted_reads),
            "nb_contradict": int(nb_contradict),
            "base_cover_left": int(base_cover_left),
            "base_cover_right": int(base_cover_right)
        }



# -*- coding: utf-8 -*-
"""Classes and functions for reading/writing STAR log."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.0.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'


class STARLog:
    """
    Reader of log file produced by STAR.

    :Example of content:
        .. highlight:: python
        .. code-block:: python

                                         Started job on |       Jan 06 13:04:49
                                     Started mapping on |       Jan 06 13:06:46
                                            Finished on |       Jan 06 13:10:10
               Mapping speed, Million of reads per hour |       237.83

                                  Number of input reads |       13477055
                              Average input read length |       150
                                            UNIQUE READS:
                           Uniquely mapped reads number |       7988755
                                Uniquely mapped reads % |       59.28%
                                  Average mapped length |       150.47
                               Number of splices: Total |       2900295
                    Number of splices: Annotated (sjdb) |       2866020
                               Number of splices: GT/AG |       2870546
                               Number of splices: GC/AG |       22895
                               Number of splices: AT/AC |       2759
                       Number of splices: Non-canonical |       4095
                              Mismatch rate per base, % |       0.26%
                                 Deletion rate per base |       0.01%
                                Deletion average length |       1.36
                                Insertion rate per base |       0.00%
                               Insertion average length |       1.41
                                     MULTI-MAPPING READS:
                Number of reads mapped to multiple loci |       4957483
                     % of reads mapped to multiple loci |       36.78%
                Number of reads mapped to too many loci |       23143
                     % of reads mapped to too many loci |       0.17%
                                          UNMAPPED READS:
               % of reads unmapped: too many mismatches |       0.00%
                         % of reads unmapped: too short |       3.72%
                             % of reads unmapped: other |       0.05%
    """

    def __init__(self, path):
        """
        Build and return an instance of STARLog.

        :param path: Path to STAR log file.
        :type path: str
        :return: The new instance.
        :rtype: STARLog
        """
        self.filepath = path
        self.input = {"nb": 0, "average_length": 0}
        self.unique = {
            "nb": 0,  # Uniquely mapped reads number |       7988755
            "average_mapped_length": 0,  # Average mapped length |       150.47
            "nb_spliced": 0,  # Number of splices: Total |       2900295
            "nb_spliced_annotated": 0,  # Number of splices: Annotated (sjdb) |       2866020
            "nb_spliced_non_canonical": 0,  # Number of splices: Non-canonical |       4095
        }
        self.multi = {
            "nb": 0,  # Number of reads mapped to multiple loci |       4957483       + Number of reads mapped to too many loci |       23143
            "nb_too_many": 0  # Number of reads mapped to too many loci |       23143
        }
        self.unmapped = {
            "nb": 0
        }
        self._parse()

    def _parse(self):
        with open(self.filepath) as FH_log:
            for line in FH_log:
                fields = [field.strip() for field in line.split("|")]
                if len(fields) == 2:
                    key = fields[0]
                    value = fields[1]
                    if key.startswith("Number of input reads"):
                        self.input["nb"] = int(value)
                    elif key.startswith("Average input read length"):
                        self.input["average_length"] = int(value)
                    elif key.startswith("Uniquely mapped reads number"):
                        self.unique["nb"] = int(value)
                    elif key.startswith("Number of reads mapped to multiple loci"):
                        self.multi["nb"] += int(value)
                    elif key.startswith("Number of reads mapped to too many loci"):
                        self.multi["nb"] += int(value)
        self.unmapped["nb"] = self.input["nb"] - self.unique["nb"] - self.multi["nb"]
